<?php if (!defined('ROOTPATH')) exit('No direct script access allowed'); ?>
<?php

require_once APPPATH . 'base/controllers/admin.php';

/**
 * User Admin Controller
 *
 * Contains all admin actions to view, edit and add users and shows an
 * overview of available users and roles.
 */
final class Users_controller extends AdminController
{
	public function _init()
	{
		parent::_init();
		$this->load->model('user');
		$this->load->model('group');
		$this->load->model('role');
	}
	
	/**
	 * Index
	 *
	 * Shows a list of users.
	 */
	public function overview($show)
	{
		// Load all/active users, groups and roles
		$users = $this->user_model->get_all_with_roles($show == 'all');
		$groups = $this->group_model->get_all_with_user_count();
		$roles = $this->role_model->get_all();
		
		// Find the active and inactive users
		$active_count = $this->user_model->get_active_count();
		$inactive_count = $this->user_model->get_inactive_count();

		$params['users'] = $users;
		$params['roles'] = $roles;
		$params['groups'] = $groups;
		$params['active_count'] = $active_count;
		$params['inactive_count'] = $inactive_count;
		$params['show'] = $show;
		$params['own_user_id'] = $this->auth->user_id();
        $params['tab'] = $this->session->get('users_tab') ?: 1;
		$this->load->page('admin/users/overview', $params);
        $this->session->set('users_tab',1);
	}

	private function _check_email($email, $user_id = null)
	{
		$in_use = $this->user_model->is_email_in_use(
			$email, 
			$user_id);

		if ($in_use)
		{
			$this->validation->add_error(
				lang('users_error_emailinuse'), 
				'email'
			);
			
			return false;
		}
		else 
		{
			return true;
		}
	}
	
	private function _check_password($arguments)
	{
		$password = $arguments['password'];

		if (!$password)
		{
			$this->validation->add_error(
				langf('validation_required', lang('users_password')),
				'password'
			);

			return false;
		}
		else
		{
			$confirm = $arguments['confirm'];
			if (!$confirm)
			{
				$this->validation->add_error(
					langf('validation_required', lang('users_confirm')),
					'confirm'
				);

				return false;
			}
			else
			{
				if ($confirm != $password)
				{
					$this->validation->add_error(
						lang('users_password_mismatch'),
						'confirm'
					);

					return false;
				}
			}
		}

		// Check if there is any password policy configured that we
		// must match.
		if (!passwords::check_from_settings($password, $policy))
		{
			$policy_desc = arr::get($policy, 'description');

			if ($policy_desc)
			{
				$this->validation->add_error(
					langf(
						'users_password_policy_mismatch_fmt',
						nl2br(h($policy_desc))
					),
					'password',
					true // Safe for HTML
				);
			}
			else 
			{
				$this->validation->add_error(
					lang('users_password_policy_mismatch_admin'),
					'password'
				);
			}

			return false;
		}

		return true;
	}
	
	/**
	 * Add
	 *
	 * Allows an administrator to add a new user.
	 */	
	public function add()
	{
		$license = license::get_current();
		
		// Check for a valid license and if the admin can add new users.
		if (!$license->is_valid)
		{
			url::redirect('admin/license');
		}
		else
		{
			if ($license->users == $license->seats)
			{
				url::redirect_error('admin/users/overview',
					langf('settings_license_error_nonewuser', 
						$license->users,
						$license->seats,
						url::site('admin/license')
					)
				);
			}
		}

		$invite_enabled = $this->settings->get_bool('invite_users',	
			true);

		$user = $this->auth->user();
		
		if (request::is_post())
		{
			if ($this->_validate($arguments))
			{
				$this->_check_email($arguments['email']);
				
				if (!$this->validation->has_error())
				{
					$invite = $arguments['invite'];
					
					if (!$invite)
					{
						$this->_check_password($arguments);
					}
					else 
					{
						if (!$invite_enabled)
						{
							$this->validation->add_error(
								lang('users_invite_disabled')
							);							
						}
					}
				}
					
				if (!$this->validation->has_error())
				{
					$user_id = $this->user_model->insert_ex(
						$arguments['name'],
						$arguments['email'],
						$arguments['language'],
						$arguments['locale'],
						$arguments['timezone'],
						$arguments['is_active'],
						$arguments['is_admin'],
						$arguments['notifications'],
						$invite ? null : $arguments['password'],
						$arguments['role_id'],
						$arguments['group_ids'],
						$password);

					if (goals::is_enabled())
					{
						goals::set_checked(TP_GOALS_USER);
					}

					if ($invite)
					{
						try 
						{
							$this->_send_invite($user_id);
							url::redirect_success(
								'admin/users/overview',
								lang('users_success_add_invite')
							);
						}
						catch (Exception $e)
						{
							url::redirect_error('admin/users/overview',
								langf(
									'users_error_add_invite',
									$arguments['email']
								)
							);
						}
					}
					else 
					{
						url::redirect_success(
							'admin/users/overview',
							lang('users_success_add')
						);
					}
				}
			}
		}
		else 
		{
			$this->validation->set_default('is_active', true);
			$this->validation->set_default('invite', true);
			$this->validation->set_default('notifications', true);
			
			$default_role = $this->role_model->get_default();
			if ($default_role)
			{
				$this->validation->set_default('role_id', 
					$default_role->id);
			}
		}
		
		$params['return_url'] = $this->referrer->return_url();
		$params['target_url'] = $this->referrer->target_url();
		$params['languages'] = $this->i18n->get_languages();
		$params['locales'] = $this->i18n->get_locales();
		$params['timezone'] = $this->i18n->get_timezones();
		$params['roles'] = $this->role_model->get_all();
		$params['groups'] = $this->group_model->get_all_with_user_count();
		$params['show_invite'] = $invite_enabled;
        $this->session->set('users_tab',1);
		$this->load->page('admin/users/add', $params);
	}
	 
	private function _send_invite($user_id)
	{
		$user = $this->user_model->get($user_id);
		
		if (!$user)
		{
			return false;
		}

		$reset = security::get_token_pair($series, $token);

		// Insert the reset token for the user and send the set/reset
		// password email.
		$this->user_model->insert_token(
			$user->id,
			TP_USER_TOKENS_RESET_PASSWORD,
			$series,
			$token,
			null, // Unnamed
			TP_USER_TOKENS_INVITE_EXPIRES
		);

		email::send_from_template(
			$user->email,
			'email_invite_user',
			array(
				'name' => $user->name,
				'login' => $user->email,
				'url' => url::app(
					str::format(
						'auth/reset_password/{0}&t={1}',
						$user->id,
						$reset
					),
					installation::get_url()
				),
				'installation_name' => $this->settings->get(
					'installation_name',
					lang('product_installation_name')
				)
			)
		);
	}
	 
	private function _check_admin($arguments, $user)
	{
		if (!$user->is_admin || !$user->is_active)
		{
			return true;
		}

		// Check if this user is the only admin and add an error if he
		// should be deactivated/degraded to a normal user.
		
		$count = $this->user_model->get_admin_count();
		
		if ($count <= 1)
		{
			if (!$arguments['is_admin'])
			{
				$this->validation->add_error(
					lang('users_error_admin_noadmin'), 
					'is_admin'
				);
				
				return false;
			}
			elseif (!$arguments['is_active'])
			{
				$this->validation->add_error(
					lang('users_error_active_noadmin'), 
					'is_active'
				);
				
				return false;
			}
		}
		
		return true;
	}
	 
	/**
	 * Edit
	 *
	 * Allows an administrator to edit an existing user.
	 */
	public function edit($user)
	{
		if (request::is_post())
		{
			if ($this->_validate($arguments))
			{
				$this->_check_email($arguments['email'], $user->id);
				
				if (!$this->validation->has_error())
				{				
					if ($arguments['password'])
					{
						$this->_check_password($arguments);
					}
				}
								
				if (!$this->validation->has_error())
				{
					$license = license::get_current();
					
					// Should the user be activated? If so, check the license.
					if ($arguments['is_active'] && !$user->is_active)
					{
						if (!$license->is_valid)
						{
							$this->validation->add_error(
								lang('users_error_active_novalidlicense'),
								'is_active'
							);
						}
						else 
						{
							if ($license->users == $license->seats)
							{
								$this->validation->add_error(
									langf('users_error_active_toomanyusers',
										$license->users,
										$license->seats,
										url::site('admin/license')
									),
									'is_active'
								);
							}
						}
					}		
				}
				
				if (!$this->validation->has_error())
				{
					$this->_check_admin($arguments, $user);
				}

				if (!$this->validation->has_error())
				{
					$this->user_model->update(
						$user->id,
						$arguments['name'],
						$arguments['email'],
						$arguments['language'],
						$arguments['locale'],
						$arguments['timezone'],
						$arguments['is_active'],
						$arguments['is_admin'],
						$arguments['notifications'],
						$arguments['password'],
						$arguments['role_id'],
						$arguments['group_ids']
					);
						
					url::redirect_success('admin/users/overview',
						lang('users_success_update'));
				}
			}
		}
		else 
		{
			$this->validation->set_defaults($user);
			$this->validation->set_default(
				'group_ids',
				$this->user_model->get_group_memberships($user->id)
			);
		}
		
		$params['languages'] = $this->i18n->get_languages();
		$params['locales'] = $this->i18n->get_locales();		
		$params['timezones'] = $this->i18n->get_timezones();
		$params['roles'] = $this->role_model->get_all();
		$params['groups'] = $this->group_model->get_all_with_user_count();
		$params['user'] = $user;
		$params['own_user_id'] = $this->auth->user_id();
		$params['external_auth_available'] = file_exists(
			path::fix(CUSPATH . 'auth/auth.php'));
		$this->load->page('admin/users/edit', $params);
	}
	
	/**
	 * AJAX: Set Role
	 *
	 * Sets the role of the given user to the specified role.
	 */
	public function ajax_set_role($user, $role)
	{
		$this->user_model->update_role($user->id, $role->id);
		ajax::return_success();
	}

	/**
	 * AJAX: Set Active
	 *
	 * Sets the active status of the given user to the specified.
	 */
	public function ajax_set_active($user, $is_active)
	{
		$this->user_model->update_active($user->id, $is_active);
		ajax::return_success();
	}

	/**
	 * Add Many
	 *
	 * Allows an administrator to add multiple users on a single page.
	 */
	public function add_many()
	{
		$default_role = $this->role_model->get_default();
		if ($default_role)
		{
			$this->validation->set_default('role_id', 
				$default_role->id);
		}

		$this->validation->set_default('invite', true);

		$params['languages'] = $this->i18n->get_languages();
		$params['locales'] = $this->i18n->get_locales();		
		$params['timezones'] = $this->i18n->get_timezones();
		$params['roles'] = $this->role_model->get_all();
		$params['groups'] = $this->group_model->get_all_with_user_count();
		$this->load->page('admin/users/add_many', $params);
	}

	/**
	 * AJAX: Render Users Preview
	 *
	 * Renders the preview for the add-many user form.
	 */
	public function ajax_render_users_preview($text)
	{
		$users = $this->_parse_and_validate_users($text);

		ajax::return_view(
			'admin/users/add_many/preview',
			array(
				'users' => $users
			)
		);
	}

	/**
	 * AJAX: Render Users Dialog
	 *
	 * Renders the action dialog for the add-many user form.
	 */
	public function ajax_render_users_dialog($text)
	{
		if (!$text)
		{
			$this->_raiset('users_users_invalid_empty');
		}

		$users = $this->_parse_and_validate_users($text, false);

		if (!$users)
		{
			$this->_raiset('users_users_invalid_empty');
		}

		foreach ($users as $user)
		{
			if (!$user)
			{
				$this->_raiset('users_users_invalid_syntax_long');
			}
		}

		ajax::return_success(
			array(
				'users' => $users,
				'code' => $this->load->view(
					'admin/users/add_many/dialog_content',
					array(
						'users' => $users
					),
					true
				)
			)
		);
	}

	private function _parse_and_validate_users($text,
		$include_existing = true)
	{
		$lines = text::split($text);

		// Parse the list of users into an array of user objects with
		// email address and name.
		$users_new = array();
		foreach ($lines as $line)
		{
			$columns = str::split($line, ',');

			// If there's an syntax error on the line (e.g. invalid
			// column, empty name or invalid email), we simply ignore it
			// here and add an "empty" object to indicate an error on
			// this line (which the UI then displays on the appropriate
			// line as part of the table).
			$user = null;

			if (count($columns) == 2)
			{
				$name = str::trim($columns[0]);
				if ($name)
				{
					$email = str::to_lower(str::trim($columns[1]));
					if ($email && check::email($email))
					{
						$user = obj::create();
						$user->name = $name;
						$user->email = $email;
					}
				}
			}

			$users_new[] = $user;
		}

		// Get the current list of users in this installation and build
		// a lookup table (email address -> user).
		$users_old = $this->user_model->get_all(true);
		$users_old_lookup = array();
		foreach ($users_old as $user)
		{
			$users_old_lookup[str::to_lower($user->email)] = $user;
		}

		// Check if some of the users in the text list already exist and
		// attach the ID to the user record in this case. Depending on
		// the arguments, we also exclude the existing users from the
		// result.
		$users_return = array();
		foreach ($users_new as $user_new)
		{
			$add_user = true;

			if ($user_new)
			{
				$user_old = 
					arr::get($users_old_lookup, $user_new->email);

				if ($user_old)
				{
					$user_new->id = $user_old->id;
					if (!$include_existing)
					{
						$add_user = false;
					}
				}
				else 
				{
					$user_new->id = null;
				}
			}

			if ($add_user)
			{
				$users_return[] = $user_new;
			}
		}

		return $users_return;
	}

	/**
	 * AJAX: Add
	 *
	 * Adds a new user to the TestRail installation (used by the mass-
	 * add feature).
	 */
	public function ajax_add()
	{
		$license = license::get_current();
		
		// Check for a valid license and if the admin can add new users.
		if (!$license->is_valid)
		{
			ajax::return_error(
				langf(
					'users_users_dialog_license_generic',
					$license->error
				),
				true
			);
		}
		else
		{
			if ($license->users == $license->seats)
			{
				ajax::return_error(
					langf(
						'users_users_dialog_license_nonewuser', 
						$license->users,
						$license->seats,
						url::site('admin/license')
					),
					true
				);
			}
		}

		$this->_validate($arguments);

		// Check that the email address is still available.
		if ($this->user_model->is_email_in_use($arguments['email']))
		{
			ajax::return_error(
				lang('users_error_emailinuse'), 
				true
			);
		}

		// Add the user to the database and send an email with the login
		// details.
		$user_id = $this->user_model->insert_ex(
			$arguments['name'],
			$arguments['email'],
			$arguments['language'],
			$arguments['locale'],
			$arguments['timezone'],
			true,					// Is active
			false,					// Is admin
			true,					// Notifications
			null,					// Password (auto-generate)
			$arguments['role_id'],
			$arguments['group_ids'],
			$password
		);

		if ($arguments['invite'])
		{
			try 
			{
				$this->_send_invite($user_id);
			}
			catch (Exception $e)
			{
				ajax::return_error(
					lang('users_error_add_invite_nopass'),
					true
				);
			}
		}

		if (goals::is_enabled())
		{
			goals::set_checked(TP_GOALS_USER);
		}

		ajax::return_success();
	}

    /**
     * Get columns to export ( $column => $label )
     * @return array
     */
    private function _get_columns_for_export()
    {
        $columns = array(
            'name'      => lang('users_user'),
            'email'     => lang('users_email'),
            'is_active' => lang('layout_grid_status'),
            'role_name' => lang('layout_grid_role'),
        );
        return $columns;
    }

    /**
     * Exports users into the different available formats (CSV/Excel).
     * @param string $format
     */
    private function _export($format = 'csv')
    {
        $users = $this->user_model->get_all_with_roles();
        $columns = $this->_get_columns_for_export();

        $this->load->view_final(
            'admin/users/export/' . $format,
            array(
                'filename' => path::clean_filename('users'),
                'columns' => $columns,
                'users' => $users,
                'separator' => ','
            )
        );
    }

    /**
     * Exports users into CSV format.
     */
    public function export_csv()
    {
        $this->_export('csv');
    }

    /**
     * Exports users into Excel format.
     */
    public function export_excel()
    {
        $this->_export('excel');
    }
}
